# Kivételek #

Eddigi kivételkezelés, amit láthattatok:

# C #

Visszatérési értéken keresztül. Hátránya: könnyen elfelejti az ember lekezelni, nem is gondol rá, mi van, ha függvény visszatérési értékét kell ellenőrizni (`null` esetleg még játszhat, viszont több problémát felfedezni nehézkes).

Példa:

	if (malloc(...)) {
		if (arraycpy(...)) {
			// Ha nem sikerult, nem kell, csak itt:
			if (free(..)) {
				...
			} else {
				// Free failed...
			}
		} else {
			// Arraycopy failed...
		}
	} else {
		// Malloc failed...
	}

Összeflyik a hibakezelő és tényleges kód, kusza, nehéz átlátni, lokális változók hatóköre értelmezhetetlen, etc.

> **Megjegyzés** A `setjmp()/longjmp()` hasznos, de még mindig nehézkes.

## C++ ##

Van `try-catch` blokk, ilyet elvileg láttatok:

* mindent lehet dobálni,
* specifikálni a tényt, hogy egy függvény hibát dob, nem kötelező,
* lekezelni a hibát nem kötelező

## Java ##

Van `try-catch-finally` blokk, kicsit strictebb szabályozás.
Általános forma:

``` java
try {
	... // Kritikus utasitasok
} catch (Exception1 e1) {
	...
} catch (Exception2 e2) {
	...
} finally {
	...
}
```

Ahol:

* `try`: kritikus utasítások egységbe zárása
* `catch`: kivétel-lekezelő ágak, specifikusak
* `finally`: minden esetben lefut (akkor is, ha volt kivétel, akkor is, ha nem). Praktikusan objektumok lezárására kiváló (lehetne...).

A finally opcionális, de elképzelhető csak `try-catch`, `try-finally` blokk is:

``` java
try {
	...
} finally {
	...
}
	
try {
	....
} catch (Throwable t) {
	....
}
```

Az első ág, amelybe a kivétel osztályhierarchia szerint beleillik, lekezeli.
Újradobás lehetséges: `throw t;`, stb.

A lényeg: a tényleges kódot elválasztani az egyes hibákat kezelő kódtól, a fenti C-s példa pl. valahogy így nézhetne ki Javaban (*már ha lennének ilyen utasítások, de ugye nincsenek, ezt mindenki tudja?*):

``` java
try {
	malloc(...)
	arraycpy(...)
	free(...)
} catch (MallocFailedException e) {
	...
} catch (ArrayCopyFailedException e) {
	...
} catch (FreeFailedException e) {
	...
}
```

Mit **nem** csinálunk kivételkezelésekkel?

``` java
try {
	malloc(...)
} catch (MallocFailedException e) {
	...
}
...
try {
	arraycpy(...)
} catch (ArrayCopyFailedException e) {
	...
}
...
try {
	free(...)
} catch (FreeFailedException e) {
	...
}
```

Ez ugyanolyan átláthatatlan, katyvasz kódhoz vezet, zavaros függőségekkel, mint a fenti C-s példa. Kerülendő!

> **Megjegyzés** *Kivételes esetekben* előfordulhat a fenti konstrukció, pl. több kritikus objektumot kell mindenképpen lezárni. Azért figyeljünk oda!

Hogyan kezelünk le egy kivételt? **Üres kivétellekezelő blokkot SOHA nem írunk!** Te fogod debuggolni 2 napig a lenyelt hibát, nem én :P Amit minimum érdemes megtenni, az valami ilyesmi:

``` java
try {
	...
} catch (Exception e) {
	// Vagy kiirod a stack trace-t (ajanlott megoldas)
	e.printStackTrace();

	// Vagy csak kiirod a hibaüzenetet a standard errorra
	System.err.println(e.getMessage());

	// Vagy logolod, ha van loggered
	logger.log(e);
}
```

Alapvetően három típusú kivétel:

1. Felügyelt kivételek: definiálni kell őket a függvényben, és ha definiáltak,
le is kell őket kezelni (ősosztály: `java.lang.Exception`}, pl.
`java.lang.ClassNotFoundException`, `java.io.IOException`)
2. Nem felügyelt kivételek: nem kötelező sem difiniálni, sem lekezelni őket
(ősosztály: `java.lang.RuntimeException`, pl. `ArrayIndexOutOfBoundsException`,
`NumberFormatException`, `DivisionByZeroException`)
3. Léteznek még `Error`-ok, ezek a `Throwable` leszármazottai. Kritikus esetben
fordulnak elő, a legtöbb esetben a lekezelésük is felesleges (pl.
`OutOfMemoryError`, `StackOverflowError`)

A hierarchia tehát valami ilyesmi:

* Throwable
	* Exception (*Felügyelt kivételek*)
		* IOException
		* FileNotFoundException
		* ...
		* RuntimeException (*Nem felügyelt kivételek*)
			* ArrayIndexOutOfBoundsException
			* NumberFormatException
			* DivisionByZeroException
			* IllegalStateException
			* IllegalArgumentException
			* ...
	* Error (*Kritikus hibák*)
		* OutOfMemoryError
		* StackOverflowError

![Kivételek Osztályhierarchiája](https://github.com/rlegendi/ELTE-javagyak/raw/master/07-io/throwable_hierarchy.png "Kivételek Osztályhierarchiája")

Mindkettő őse a `Throwable` osztály, ezt kell hát lekezelni, ha mindent
lehetőségre fel akarunk készülni. **Ökölszabály:** üres kivételkezelő blokkot
**soha** ne készítsünk! Legegyszerűbb megoldás: `e.getMessage()` vagy
`e.printStackTrace()`. További funkciók az `Exception` osztály leírásában
(`javadoc`). Kivételt dobni a `throw` utasítással lehet (_mindig_ van
paramétere!).

## Példa ##
### Egyszerű kivételkezelés ###

``` java
public static void main(String[] args) {
	try {
		int res = Integer.parseInt(args[0]);
		// ...
	} catch (NumberFormatException nfe) {
		System.err.println("Hibas input: " + args[0]);
		nfe.printStackTrace();
	}
}
```

> **Megjegyzés** Kivételek neve általában `e`, de igazából ízlés kérdése.

### Előfeltétel ellenőrzés ###

Lehetőleg minél hamarabb, a függvény elején - ha sérül az előfeltétel, dobj `IllegalArgumentException` kivételt (nem felügyelt, így deklarálni sem kell).

``` java
public static int minPlace(final int[] arr) {
	if ( 0 == arr.length ) {
		throw new IllegalArgumentException("Empty array.");
	}

	// Search for the index of the minimal element in the array
}
```


### Függvénydefiníció ###
Ha ki akarod kényszeríteni a kivétel lekezelését, valami beépített (vagy saját, ld. alább a példában) `Exception` leszármazottat kell használnod.

``` java
// Egyszerubb forma, ha nem akartok uzeneteket
//class ZeroParameterException extends Exception {}
	
class ZeroParameterException extends Exception {
	public ZeroParameterException() {
		super();
	}

	public ZeroParameterException(final String msg) {
		super(msg);
	}
}

public class Main {	
	public static double divide(int a, int b) throws ZeroParameterException {
		if (0 == b) {
			throw new ZeroParameterException("b erteke nem lehet 0!");
		}

		return (double) a / b;
	}
		
	public static void main(String[] args) {
		try {
			double res = divide(1, 0);
		} catch (ZeroParameterException e) {
			System.err.println(e.getMessage());
		} catch (Exception e) {
			System.err.println(e.getMessage());
		} finally {
			System.err.println("vege");
		}
	}
}
```

> **Részletesen**
>
> <http://download.oracle.com/javase/tutorial/essential/exceptions/>
>
> <http://download.oracle.com/javase/6/docs/api/java/lang/package-summary.html> (*A leggyakoribb kivételeket megnézegetni*)

Saját kivétel is definiálható, csak származtatni kell (pl. a
`java.lang.RuntimeException`, `java.lang.Exception` osztályokból, ezekről
később).

> **Megjegyzés** Az alábbi *Explanations of Common Java Exceptions* egy humoros
> leírása a common Java kivételeknek. Mielőtt komolyan vennétek, hangsúlyozom,
> ez csak poén, de ha már megragad 1-2 Exception neve, akkor már megérte
> átröhögni :-) Thx to Björn Andersson!

# Streamek #
Csatornák, absztrakciós szint bemenet-kimenet kezelésére (`hiext`, `lorem`).

> **Részletesen:** <http://download.oracle.com/javase/tutorial/essential/io/>

## Példa ##
``` java
package io;
	
import java.io.FileWriter;
import java.io.PrintWriter;
	
public class WriteSampleFile {
    public static void main(String[] args) throws Exception {
        PrintWriter pw = new PrintWriter( new FileWriter("dummy.txt") );
        pw.println("Dummy data here");
        pw.close();
    }
}
```

## Csoportosítás ##

Szervezés szerint:

* `InputStream`, `OutputStream` (bájtszervezésű)
* `Reader`, `Writer` (karakterszervezésű)

Ezekből rengeteg változat, pl.

* `StringReader`, `FileReader`, `BufferedReader`, etc.
* `FileInputStream`, `DataInputStream`, etc.
* `FileOutputStream`, `PrintStream`, etc.
* `FileWriter`, `PrintWriter`, etc.

Pl. `FileReader`, `FileWriter`, `FileInputStream`, `FileOutputStream`

Feladat szerinti csoportosítás:

* Adatforrás, adatnyelő (pl. `FileInputStream`, `FileReader`)
* Szűrők: meglévő csatornához plusz funkcionalitás (pl. `BufferedInputStream`).
  Összekapcsolhatók (pl. `SequenceInputStream`), csatornák hozhatók létre (pl.
  kommunikációs közeg szálaknak, `PipedInputStream`, stb.)

## Alapvető műveletek ##
* Megnyitás automatikus, lezárás: `close()` (ne felejtsd el, mert nem fogsz
  tudni hozzáférni)
* `flush()`: automatikus endline karakterre a szöveges fájlokban.
* Kiírás: `write()`, `print()`
* Olvasás: `read()`. Ha a csatorna kiürül, akkor az olvasó művelet
  *blokkolódik*, amíg adatot nem kap:

	``` java
	public static void main(String[] args) throws Exception {
		int i = System.in.read();
		System.out.println("A kapott karakter: " + i);
	}
	```

* *"Könyvjelző-mechanizmus"*, ha támogatott: `markSupported()`, `mark()`,
  `reset()`
* Padding átugrása: `skip()`
* Csatorna ürességének ellenőrzése: `ready()`
* Hány bájtot lehet minimum olvasni? `available()`

	``` java
	int meret = new FileInputStream("tmp.txt").available();
	```

## Speciális streamek ##

* Sorok számolására: `LineNumberInputStream#getLineNumber()`
* Adatok olvasása, visszafűzése a csatornára: `PushbackInputStream`,
  `PushbackReader`
* Bufferelt csatornák: `BufferedReader`, `BufferedInputStream`
* Véletlen elérésű fájlok: `RandomAccessFile` (későbbb kicsit részletesebben is)
* Szövegfeldolgozás: `StringTokenizer` (van sima `StreamTokenizer` is)
* `System.in`, `out`, `err`: `InputStream` ill. `PrintStream`-ek. Pl. standard
  input átirányítása:

	``` java
	BufferedReader br = new BufferedReader( new InputStreamReader(System.in) );
	```

> **Részletesen:** <http://download.oracle.com/javase/6/docs/api/java/io/package-summary.html>

## Példák ##

### Fájl írása ###

``` java
package files;
	
import java.io.FileNotFoundException;
import java.io.PrintWriter;
	
public class WriteFile {
	public static void main(String[] args) {
		PrintWriter pw = null;
	        
		try {
			pw = new PrintWriter(args[0]);

			pw.println("Line1");
			pw.println("Line2");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} finally {
			if (pw != null) pw.close();
		}
	}
}
```

### Fájl olvasása ###

``` java
package files;
	
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
	
public class ReadFile {
	public static void main(String[] args) {
		BufferedReader br = null;
	        
		try {
			br = new BufferedReader(new FileReader(args[0]));
			String line = null;
			while ((line = br.readLine()) != null) {
				System.out.println(line);
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (br != null) {
				try {
					br.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
	}
}
```

## Sor szavakra bontása ##
Megoldás 1:

``` java
String[] words = line.split(" ");
```

Nem jó, mi van, ha a tényleges szöveg előtt-után szóköz van? Így pl. `" a b"` darabolásából egy három elemű tömböt kapunk, amiben az első elem az üres string: `{"", "a", "b"}`.

``` java
String[] words = line.trim().split(" ");
```

Nem jó, mert mi van, ha tabulátor karakter?

Megoldás 2: `String#split()` alatt elérhető [regexp patternek](http://docs.oracle.com/javase/6/docs/api/java/util/regex/Pattern.html#sum) tanulmányozása után:

``` java
String[] words = line.trim().split("\\s")
```

Nem jó, mert pl. ha két szóköz van egymás mellett, akkor megint üres elemeket kapok két értékes szó közé. Megoldás: default greedy mintaillesztés:

``` java
String[] words = line.trim().split("\\s+")
```

Ez már majdnem mindenre jó lesz :-)

## Egyéb megjegyzések ##

* Oprendszer függő tulajdonságok:

	``` java
	// PATH elvalaszto, pl. ":" - Unix, ";" - Windows
	final String PATH_SEPARATOR = File.pathSeparator;
	// Ugyanaz, csak karakterkent
	final char PATH_CHAR = File.pathSeparatorChar;
			
	// Nev szeparator, pl. "/" - Unix, "\" - Windows
	final String SEPARATOR = File.separator;
	// Ugyanaz, csak karakterkent
	final char SEPARATOR_CHAR = File.separatorChar;
			
	// Sorvege karakter, pl. "\n" - Unix, "\r\n" - Windows
	// Reszletesen lasd FAQ!
	final String EOL = System.getProperty("line.separator");
	```

* `FileNotFoundException`: Ha nem tudod, hol keresi az IDE, vagy a Java a
  megadott file-t, alkalmazhatod ezt a cselt:

	``` java
	File file = new File("test.txt");
	System.out.println( file.getAbsolutePath() );
	```

### Java 7 újdonságai ###

A Java 7 tartalmaz egy teljesen újraírt I/O könyvtárat is, amely rendelkezik pl. a következő hasznos függvénnyel is:

``` java
Path file = Paths.get(filename);
List<String> lines = Files.readAllLines(file, Charset.defaultCharset());
```

> **Részletesen** Rengeteg hasznos segédfüggvény:
>
> http://docs.oracle.com/javase/7/docs/api/java/nio/file/Files.html
>
> http://docs.oracle.com/javase/tutorial/essential/io/fileio.html

#### A try-with-resources utasítás ####

``` java
package files;
 
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
 
public class ReadFileJava7 {
	public static void main(String[] args) {
		try (BufferedReader br = new BufferedReader(new FileReader(args[0]))) {
			String line = null;
 
			while ((line = br.readLine()) != null) {
				System.out.println(line);
			}
		} catch (IOException e) {
			e.printStackTrace();
		} 
	}
}

```

> **Részletesen**
>
> https://github.com/rlegendi/ELTE-javagyak/blob/master/I-java-1.7/try-with-resources.md

### Közvetlen elérésű fájlok ###
A `java.io.RandomAccessFile`, kb. mint egy bájtvektor, olvasható és írható.
Fájlmutató az aktuális pozícióra, ez lekérdezhető (`getFilePointer()`),
állítható (`seek(int)`). Implementálja mind a `DataInput`, `DataOutput`
interfészeket (mindkettőt egyszerre), a műveleteivel tetszőleges típus írható,
olvasható (úgy használható, mint a `DataInputStream`, `DataOutputStream`:
`write*()`, `read*()` függvények), byte-ok átugorhatók (`skip(int)`).

Példaalkalmazás:

``` java
package files;
		
import java.io.IOException;
import java.io.RandomAccessFile;
	
public class RandomFileTest {
	public static void main(final String[] args)
		throws IOException {
			final RandomAccessFile raf =
			new RandomAccessFile("dummy.dat", "rw");
	
			raf.writeInt(0xCAFEBABE);
			raf.seek(16);
			raf.writeInt(0xDEADBEEF);
			raf.seek(32);
			raf.writeInt(0xBADF00D0);
			raf.seek(48);
			raf.writeInt(0xDEADC0DE);
			raf.close();
	}
}
```

> **Megjegyzés:** bájtokat írunk ki, nem elfelejteni! Hexa módban kell megnyitni
> a fájlt, hogy lássuk ténylegesen mi is íródott ki, pl.
> *Total Commander + F3 + 3*, vagy *vi + ":\%!xxd"*.

Gotchaz:

* `write(int)` - hiába van int paramétere, csak byte-ot ír ki, a legalját
* `seek()`, `write(...)` - nem tolódik tovább a stream, kézzel kell mindent
  odébbmásolni
* `raf.seek(file.length());` - file végére ugrás, ahhoz hozzáfűzés
* `write()` - elfogad `byte[]` paramétert, de `writeBytes()` - csak `Stringet`
* `EOFException` - elindexelésnél (`IOException` leszármazottja)

> **Részletesen:** <http://download.oracle.com/javase/6/docs/api/java/io/RandomAccessFile.html>

## Feladatok ##
* Készíts egy szűrő programot, amely két parancssori argumentumot kap: egy input
  fájl elérési utat, és egy stringet! A program olvassa be az inputfájlt, és
  írja ki a képernyőre azokat a sorokat, amelyekben a megadott karakterlánc
  előfordul!

* Készíts egy programot, amely két parancssori argumentumot kap: egy input fájl
  elérési utat, és egy stringet! A program olvassa be az inputfájlt, és számolja
  meg a megadott szó előfordulásainak számát!

* Készíts egy interaktív programot, amely a felhasználó által megadott sorokat
  (egészen az üres sorig) elmenti egy fájlba. Az output fájl nevét kérje be a
  program indulásnál.

* Készíts egy programot, amely egy parancssori argumentumként megadott fájlnak
  kiírja az első `N` sorát (`N=5` alapértelmezett értékkel rendelkezik, azonban
  opcionális  parancssori argumentummal ezt állítani lehessen).

* Készíts egy programot, amely egy parancssori argumentumként megadott fájlnak
  kiírja az utolsó `N` sorát (`N=5` alapértelmezett értékkel rendelkezik,
  azonban opcionális  parancssori argumentummal ezt állítani lehessen).

* Készíts egy programot, amely három parancssori argumentumot kap: két input
  fájl és egy output fájl elérési utat. A program fűzze össze az első két fájlt
  egymás után az output fájlba!

* Adott egy fájl a következő szerkezettel:

		# Comment sor
		# Ezeket at kell ugrani
		# A file formatuma:
		# <Nev>:<Szuletesi ev>:<honap>:<nap>
		Citad Ella:1984:10:10
		Har Monika:1956:5:6
		Git Aron:1992:11:23
		Tek Elek:2002:6:14
	
	A program olvassa be ezeket az adatokat, és tárolja egy megfelelő
  	reprezentációban (készíts pl. egy `Person` osztályt a szükséges adattagokkal).
	A program parancssori argumentumként kapjon egy hónapot, és listázza ki, hogy
	az adott hónapban mely személyeknek van születésnapja. Használjátok a `String`
	osztály `split()` metódusát.

* Készíts egy programot, amely kiírja egy parancssori argumentumként megadott
  fájl páratlan sorait a képernyőre, páros sorait pedig egy `"even.txt"` fájlba!
  Ezután a `"even.txt"` tartalmával felülírja az eredeti fájlt.

* Készítsünk egy egyszerű *l33t5p34k* (leetspeak) generátort! A program 1
  parancssori argumentumot kapjon: egy input fájl elérési utat. A program
  olvassa be az input fájlt, és minden szón végezze el a következő
  módosításokat, majd írja ki a képernyőre a módosított szöveget:
  
	* Ha a szó vége:
		* `"s"` helyett `"z"`
		* `"ck"` helyett `"x"`
	* `"a"` helyett `"@"`
	* `"e"` helyett `"3"`
	* `"i"` helyett `"1"
	* `"o"` helyett `"0"`
	* `"u"` helyett `"v"`
	* `"f"` helyett `"ph"`
	* `"s"` helyett `"$"`
	* `"g"` helyett `"9"
	* `"y"` helyett `"j"`
	* `"t"` helyett `"+"`
	* `"!"` helyett `"1"`
	* Ezen kívül minden 2. karakter esetén a kisbetűből csináljon nagyot, a
	  nagybetűből kicsit!

* Készítsünk egy Sudoku ellenőrző programot! A program inputja egy fájl,
  amelynek soraiban pontosan 81 karakter található, és egy-egy Sudoku
  sorfolytonos ábrázolását jelenti. A program outputja egy fájl legyen, amelyben
  az érvényes kitöltést tartalmazó sorok kerülnek bele!

	Példa input fájl sorok:
	
		123456789123456789123456789123456789123456789123456789123456789123456789123456789
		123456789234567891345678912456789123567891234678912345891234567891234567912345678

* Készítsünk egy egyszerű konzolos alkalmazást, amely képes fájlok bináris
  karbantartására! A program első paraméterként kapja meg a szerkesztendő fájl
  nevét. A további paraméterek a következők lehetnek:
		
	* `mb <pozíció> <byte>`: az adott pozíción lévő byte érték módosítása
	* `mi <pozíció> <int>`: az adott pozíción lévő int érték módosítása
	* `i <pozíció> <szöveg>`: a specifikált szöveg beillesztése az adott
	  pozícióra. A program illessze be a szöveget, azaz tolja el a bájtokat
	  megfelelőképp

